/*+
 *	Module:		llrecover.c
 *
 *	Purpose:	To provide FMQ error correction.
 *			This error recovery algorithm is taken from the paper
 *			Fischer, C.N., Milton, D.R., and Quiring, S.B.,
 *			"Efficient LL(1) Error Correction and Recovery Using
 *			Only Insertions", Acta Informatica 13, pp 141-154, 1980.
 *
 *			The algorithm is a slight modification of that given in
 *			paper, but is otherwise equivalent.  It uses error
 *			recovery tables produced by a specially modified version
 *			of llgen.  These are stored in external data structures
 *			initiallized in ll.tab.c, the output of llgen.
 *
 *	Invocation:	recovered = llrecover (current, input, attptr);
 *
 *	Parameters:
 *
 *	  Input:	current - The address of the current top of
 *			stack symbol.
 *
 *			input - The current input symbol.
 *
 *			attptr - The address of the attributes associated
 *			with the current top of stack symbol.
 *
 *	  Output:	recovered - The truth or falsity of the algorithm's
 *			ability to recover from the current error situation.
 *
 *	Author:		D. Bocking
 *
 *	Date:		12 December, 1984
-*/

#include	"llparam.h"
#include	"llglobal.h"
#include	<stdio.h>

/*
 *	The following structure is used to store input tokens and
 *	their associated attributes during the test parses done
 *	in attempting to recover from an error.
 */

struct	lookahead	{

	int	lk_token;
	int	lk_line;
	int	lk_nattr;
	int	*lk_attrs;
};

struct	lookahead	pushback [MAXCONTEXT + 1];
int	npushback	= 0;

int	pushattr [(MAXCONTEXT + 1) * VSIZE];
int	npushattr = 0;

/*
 *	The following structure and its typedef define the structure
 *	of the substring tables used by the recovery algorithm to
 *	calculate the strings of tokens to be inserted during error
 *	recovery.
 */

#define	SUBSTRMAX	25

struct	strtab	{

	int	s_num;			/* The number of active substrings */
	short	*s_str [SUBSTRMAX];	/* Pointers to the substrings*/
};

/* typedef	strtab	STRTAB; */

short	llnullstr [] = {
	-1				/* A null string */
};

#define	UNDEFINED	(short *) (-1)	/* An undefined string */
#define	INFINITY	16383		/* The cost of an undefined string */

/*
 *	The main recovery algorithm.
 */

llrecover (current, input, attptr)

int	*current;
int	input;
int	*attptr;

{
	register	int	insym;
	register	int	i;
	struct	strtab	insert;
	int	recovered;	/* true on successful recovery */
	short	*sptr;
	char	str [256], *llprtoken ();

	if (lldebug & DBGSTP1) {
		fprintf (stderr, "Attempt error recovery\n");
		fprintf (stderr, "\tTop of stack: %s\n", symbols [*current]);
		fprintf (stderr, "\tInput symbol: %s\n", symbols [input]);
	}

	/*
	 *	Calculate the token string to be inserted.
	 */

	llmkinsert (current, input, attptr, &insert);

	if (insert.s_str [0] != (short *) (-1)) {
		recovered = 1;

		/*
		 *	Push back the original input symbol.
		 *	Then, push back the inserted tokens in
		 *	in reverse order, since the pushback
		 *	mechanism is a stack.
		 */

		llpushback (input, llline, attptr);

		for (i = insert.s_num - 1; i >= 0; i--) {
			sptr = insert.s_str [i];
			if (sptr != UNDEFINED) {
				llpushstrg (sptr);
			}
		}

		/*
		 *	Notify the rest of the compiler that a string
		 *	of tokens is being inserted to correct a
		 *	syntax error.
		 */

		str [0] = '\0';
		for (i = npushback - 1; i > 0; i--) {
			strcat (str, " ");
			strcat (str, llprtoken (pushback [i].lk_token,
			pushback [i].lk_attrs));
		}

		lladvise (1, llline, input, attptr, str);
	} else {
		recovered = 0;
		lladvise (2, llline, input, attptr, "");
	}
	return (recovered);
}

/*
 *	The following subroutine constructs the string of tokens
 *	to be inserted into the input stream to correct an error.
 */

llmkinsert (curtop, input, attrib, insert)

int	*curtop;
int	input;
int	attrib [];
struct	strtab	*insert;

{
	int	insertcost, prefixcost, scost, ecost;
	int	toptoken;
	short	*sstr, *estr, *llgetS (), *llgetE ();
	struct	strtab	prefix;

#ifdef	DEBUG
	fprintf (stderr, "Mkinsert - top: %s input: %s\n", symbols [*curtop],
	symbols [input]);
#endif
	/*
	 *	Initialize the prefix and insert string tables so
	 *	that prefix is the empty string and insert is
	 *	the undefined string.
	 */

	prefix.s_num = 1;
	prefix.s_str [0] = llnullstr;
	prefixcost = 0;

	insert -> s_num = 1;
	insert -> s_str [0] = UNDEFINED;
	insertcost = INFINITY;

	/*
	 *	Now, examine each of the terminal and nonterminal
	 *	symbols on the parse stack, to find a prefix or
	 *	error correction string.
	 */

	do {
		toptoken = *curtop--;

#ifdef	DEBUG
		fprintf (stderr, "Consider toptoken: %s\n",
		symbols [toptoken]);
#endif
		if (toptoken < llfa) {
			estr = llgetE (toptoken, input);
			ecost = llgetcost (estr);

			sstr = llgetS (toptoken);
			scost = llgetcost (sstr);

#ifdef	DEBUG
			fprintf (stderr, "InsertCost: %d PrefixCost: %d\n",
			insertcost, prefixcost);

			fprintf (stderr, "Ecost: %d Scost: %d\n", ecost,
			scost);
#endif

			if (insertcost > lladdcost (prefixcost, ecost)) {
				/* insert = prefix cat estr */
				llcpystrg (insert, &prefix);
				llcatstrg (insert, estr);

				insertcost = lladdcost (prefixcost, ecost);

#ifdef	DEBUG
				fprintf (stderr,
				"\tInsert = Prefix cat Estr\n");
#endif
				if (ecost <= scost) {
					/* Cheapest insert, return it */
					break;
				}
			}

			/* prefix = prefix cat sstr */

			llcatstrg (&prefix, sstr);
			prefixcost = lladdcost (prefixcost, scost);
#ifdef	DEBUG
			fprintf (stderr, "\tPrefix = Prefix cat Sstr\n");
#endif
		}

		curtop -= llinhs [toptoken] + llsyns [toptoken];
	} while (toptoken != 0);

	/* return insert */
}

/*
 *	The following subroutines are used to locate strings in
 *	the S and E tables because large portions of these tables
 *	are constants which are not generated into ll.tab.c
 */

short	*
llgetS (symbol)

int	symbol;

{
	return (llfmqS [symbol]);
}

short	*
llgetE (sym1, sym2)

int	sym1;
int	sym2;

{
	if (sym1 < llfnt) {
		if (sym1 == sym2) {
			return (llnullstr);
		} else {
			return (UNDEFINED);
		}
	}

	return (llfmqE [(sym1 - llfnt) * llfnt + sym2]);
}

/*
 *	The following subroutines are used to copy struct strtab structures and
 *	to concatenate a new string to the end of an existing struct strtab.
 */

llcpystrg (to, from)

struct	strtab	*to;
struct	strtab	*from;

{
	register	int	i;

	to -> s_num = from -> s_num;

	for (i = 0; i < from -> s_num; i++) {
		to -> s_str [i] = from -> s_str [i];
	}
}

llcatstrg (to, str)

struct	strtab	*to;
short	*str;

{
	if (to -> s_num >= SUBSTRMAX) {
		fprintf (stderr, "COMPILER ERROR: %d Error Recovery %s.\n",
		llline, "String Too Long");
		exit (1);
	}

	to -> s_str [to -> s_num++] = str;
}

/*
 *	The following subroutine is used to push a string of symbols
 *	back on the input in reverse order, since the push back storage
 *	is a stack.
 */

llpushstrg (strng)

short	*strng;

{
	int	newattr [VSIZE];

	if (*strng == -1) {
		return;
	} else {
		llpushstrg (strng + 1);
	}

	lldogensyn (*strng, newattr);
	llpushback (*strng, llline, newattr);
}

/*
 *	The following subroutine calculates the cost of a string.
 */

llgetcost (str)

short	*str;

{
	register	short	*sptr;
	register	int	cost, totalcost;

	if ((sptr = str) == UNDEFINED) {
		return (INFINITY);
	}

	totalcost = 0;
	for (sptr = str; *sptr != -1; sptr++) {
		cost = llfmqcosts [*sptr];
		if (cost == INFINITY) {
			return (cost);
		}
		totalcost += cost;
	}

	return (totalcost);
}

/*
 *	The following routine adds two string costs together, taking
 *	account of the role of INFINITY.
 */

lladdcost (cost1, cost2)

int	cost1;
int	cost2;

{
	if (cost1 == INFINITY || cost2 == INFINITY) {
		return (INFINITY);
	}

	return (cost1 + cost2);
}

/*
 *	This routine is used to pushback previously scanned or artificially
 *	inserted terminal symbols onto the terminal symbol input stream
 *	to allow for rescanning.
 */

llpushback (symbol, line, attptr)

int	symbol;
int	line;
int	*attptr;

{
	int	i;

	/*
	 *	Check to be sure the maximum pushback level has not
	 *	been exceeded.
	 */

	if (npushback >= MAXCONTEXT + 1) {
		fprintf (stderr, "COMPILER ERROR: %s: symbol %s\n",
		"Maximum pushback exceeded", symbols [symbol]);
		exit (1);
	}

	/*
	 *	Make sure the symbol is a terminal symbol.
	 */

	if (symbol >= llfnt) {
		fprintf (stderr, "Pushback of nonterminal: %s\n",
		symbols [symbol]);
		exit (0);
	}

	/*
	 *	Copy the symbol into the pushback table.
	 */

	pushback [npushback].lk_token = symbol;
	pushback [npushback].lk_line = line;
	pushback [npushback].lk_nattr = llsyns [symbol];
	pushback [npushback].lk_attrs = &pushattr [npushattr];

	for (i = 0; i < llsyns [symbol]; i++) {
		pushattr [npushattr++] = attptr [i];
	}

	npushback++;
}

/*
 *	The following routine is the second level scanner.  If there
 *	are any previously pushed back symbols, the most recently pushed
 *	symbol is popped and returned to the user.  Otherwise, the
 *	user's scanning routine is called directly.
 */

llscn2 ()

{
	int	symbol, i;
	struct	lookahead	*pbptr;

	/*
	 *	Check for a previously pushed back symbol.
	 */

	if (npushback > 0) {
		npushback--;

		pbptr = &pushback [npushback];
		symbol = pbptr -> lk_token;
		llline = pbptr -> lk_line;

		for (i = 0; i < pbptr -> lk_nattr; i++) {
			llsyn (pbptr -> lk_attrs [i]);
		}
		npushattr -= pbptr -> lk_nattr;
	} else {
		symbol = llscan ();
	}

	return (symbol);
}
