/*+
 *	Module:		llattrib.c
 *
 *	Purpose:	To define the subroutines that manipulate the attribute
 *			structure type.
 *
 *	Invocation:	clrattr (attr);
 *
 *			addattr (attr, name);
 *
 *			setattr (attr, name, value);
 *
 *			attval = getattr (attr, name);
 *
 *			trueorfalse = defattr (attr, name);
 *
 *			foreachattr (attr, routine, param);
 *
 *	Parameters:
 *
 *	  Input:	attr - An instance of the ATTRIB type which is
 *			the set of attributes to be manipulated.
 *
 *			name - A character string naming the attribute
 *			in the set which is to be affected.
 *
 *			value - An integer value to be associated with
 *			the attribute.
 *
 *			routine - The address of a subroutine to be called
 *			for each defined attribute in the specified set.
 *			The routine is called as follows:
 *
 *				routine (param, name, value),
 *
 *			where param is the parameter given to foreachattr,
 *			name is the name of the specific attribute and
 *			value is its stored value.
 *
 *			param - The parameter to be passed to the routine
 *			given above.  The type of this parameter is
 *			massaged with disdain.  It should be 'univ'.
 *
 *	  Output:	attval - The value of the named attribute in
 *			the set.
 *
 *			trueorfalse - The truth or falsity concerning
 *			whether an attribute has been defined or not.
 *
 *	Author:		D. Bocking
 *
 *	Date:		16 October, 1984
-*/

#include	<stdio.h>
#include	"llattrib.h"
#include	<ctype.h>

clrattr (attr)

ATTRIB	*attr;

{
	register	int	i;

	for (i = 0; i < ASIZE; i++) {
		attr -> attval [i] = 0;
		attr -> attdef [i] = 0;
	}
}

addattr (attr, name)

ATTRIB	*attr;
char	*name;

{
	register	int	i;

	i = numattr (name);

	if (attr -> attdef [i] != 0) {
		fprintf (stderr, "Attribute %s redefined.\n", name);
		exit (1);
	}
	attr -> attdef [i] = 1;
}

defattr (attr, name)

ATTRIB	*attr;
char	*name;

{
	register	int	i;

	i = numattr (name);

	return (attr -> attdef [i]);
}

setattr (attr, name, value)

ATTRIB	*attr;
char	*name;
int	value;

{
	register	int	i;

	i = numattr (name);

	if (attr -> attdef [i] == 0) {
		fprintf (stderr, "Undefined attribute %s\n", name);
		exit (1);
	}

	attr -> attval [i] = value;
}

getattr (attr, name)

ATTRIB	*attr;
char	*name;

{
	register	int	i;

	i = numattr (name);

	if (attr -> attdef [i] == 0) {
		fprintf (stderr, "Undefined attribute %s\n", name);
		exit (1);
	}

	return (attr -> attval [i]);
}

numattr (name)

char	*name;

{
	register	int	i;
	register	char	c;

	/*
	 *	To conform to the original version of ATS, attribute
	 *	names are restricted to be 1-character strings.
	 *	I will fix this as soon as I can!
	 */

	if (strlen (name) != 1 || !isalpha (c = name [0])) {
		fprintf (stderr, "Invalid attribute name: %s\n", name);
		exit (1);
	}

	if (isupper (c)) {
		i = c - 'A';
	} else {
		i = c - 'a';
	}

	return (i);
}

foreachattr (attr, routine, param)

ATTRIB	*attr;
int	(*routine) ();
int	param;

{
	register	int	i;
	char	name [2];

	name [1] = '\0';
	for (i = 0; i < ASIZE; i++) {
		if (attr -> attdef [i] != 0) {
			name [0] = i + 'a';
			(*routine) (param, name, attr -> attval [i]);
		}
	}
}
