/*+
 *	Module:		llcosts.c
 *
 *	Purpose:	Provides subroutines used to store and calculate
 *			the costs of strings of grammar tokens.
 *
 *	Invocation:	setcost (i, cost);
 *
 *			cost = getcost (i);
 *
 *			cost = strcost (str);
 *
 *			cost = strncost (str, len);
 *
 *	Parameters:
 *
 *	  Input:	i - The token number of a token whose cost is to
 *			be set or returned.
 *
 *			cost - The cost to be assigned to a token.
 *
 *			str - The address of the first token in a string
 *			of tokens.
 *
 *			len - The number of tokens in a given token string.
 *
 *	  Output:	cost - The cost returned for a token or string of
 *			tokens.
 *
 *	Author:		D. Bocking
 *
 *	Date:		4 December, 1984
-*/

#ifdef	DEBUG
#include	"llsymtab.h"
#include	<stdio.h>
#endif
#include	"llparam.h"
#include	"llcosts.h"
#include	"llinstr.h"

/*
 *	The vector for storing the costs of grammar symbols.
 */

int	llcostvec [MAXSIZE];

/*
 *	The following routine sets the cost of a particular grammar
 *	symbol.
 */

setcost (i, cost)

int	i;
int	cost;

{
#ifdef	DEBUG
	struct	symtab	*getsymaddr ();
#endif
	if (0 <= i && i < MAXSIZE) {
#ifdef	DEBUG
		fprintf (stderr, "Setcost %s %d\n", getsymaddr (i) -> sname,
		cost);
#endif
		llcostvec [i] = cost;
	}
}

/*
 *	The following routine returns the cost of a particular grammar
 *	symbol.
 */

getcost (i)

int	i;

{
	if (0 <= i && i < MAXSIZE) {
		return (llcostvec [i]);
	}

	return (INFINITY);
}

/*
 *	The following routine returns the cost of a string of grammar
 *	symbols of indefinite length.  The string is assumed to be
 *	terminated by a token value of -1.
 */

strcost (str)

INSTR	*str;

{
	register	int	i, j, cost;
	register	INSTR	*strp;

	strp = str;

	if (strp == UNDEFINED) {
		cost = INFINITY;
	} else {
		cost = 0;
		while ((i = *strp++) != -1) {
			if ((j = llcostvec [i]) == INFINITY) {
				cost = INFINITY;
				break;
			}

			cost += j;
		}
	}

	if (cost > INFINITY) {
		cost = INFINITY;
	}
	return (cost);
}

/*
 *	The following routine calculates the cost of a string
 *	of grammar symbols whose length is explicitly known.
 */

strncost (str, len)

int	*str;
int	len;

{
	register	int	i, j, cost, *strp;

	strp = str;

	cost = 0;
	for (i = 0; i < len; i++) {
		if ((j = llcostvec [*strp++]) == INFINITY) {
			cost = INFINITY;
			break;
		}

		cost += j;
	}

	if (cost > INFINITY) {
		cost = INFINITY;
	}
	return (cost);
}
