/*+
 *	Module:		llfixattr.c
 *
 *	Purpose:	To provide subroutines to fix up the attribute
 *			instances in the production token table as the
 *			grammar is being parsed.
 *
 *	Invocation:	fixlhsattr (lattr, strng, value);
 *
 *			fixlhssynattr (rattr, strng, value););
 *
 *			fixrhsinhattr (lattr, rattr, strng);
 *
 *			fixrhssynattr (lattr, rattr, strng);
 *
 *	Parameters:
 *
 *	  Input:	lattr - The address of the lhs attribute table.
 *
 *			rattr - The address of the rhs attribute table.
 *
 *			strng - The address of the character string containing
 *			the name of an attribute.
 *
 *			value - The integer value to be assigned to a new
 *			lhs attribute.
 *
 *	  Output:	None.
 *
 *	Author:		D. Bocking
 *
 *	Date:		30 October, 1984
-*/

#include	"llattrib.h"
#include	<stdio.h>
#include	"llglobal.h"

/*
 *	This routine is used to enter each new lhs attribute into
 *	the lhs attribute table and to assign its value.  Dummy
 *	entries are made in the production rhs table for each
 *	attribute.  These are fixed up later, as rhs inherited
 *	attributes are encountered, and when synthesized attributes
 *	are tidied up at the end of the production.
 */

fixlhsattr (lattr, strng, value)

ATTRIB	*lattr;
char	*strng;
int	value;

{
	mkrhstoken (0);

	if (defattr (lattr, strng)) {
		fprintf (stderr, "%d: %s %s.\n", line,
		"Duplicate attribute --", strng);
		seterror ();
	}

	addattr (lattr, strng);
	setattr (lattr, strng, value);
}

/*
 *	The following routine is used to fix up lhs synthesized
 *	attributes at the end of each production.  An offset
 *	is generated to the location of the last reference in
 *	the production rhs to each of these.  This is used during
 *	parsing to link the place of synthesis in one production
 *	to the place of inheritance in a containing production.
 */

fixlhssynattr (rattr, strng, value)

ATTRIB	*rattr;
char	*strng;
int	value;

{
	register	int	rvalu, offset;

	/*
	 *	Fix up synthesized lhs attributes only.
	 */

	if (value < 0) {
		if (!defattr (rattr, strng)
		|| (rvalu = getattr (rattr, strng)) == -1) {
			fprintf (stderr, "%d %s %s.\n", line,
			"Undefined attribute --", strng);
			seterror ();
		}

		/*
		 *	Alter the lhs synthesized attribute
		 *	token so that it pointes by negative
		 *	offset to the place of synthesis in
		 *	the rhs of the production, the offset
		 *	of which is the value of the rhs attribute
		 *	of the same name.
		 */

		offset = - value - 1;
		setrhstoken (offset, offset - rvalu);
	}
}

/*
 *	This routine is used to process rhs inherited attributes.
 *	The previous reference, if any, to this attribute is altered
 *	so that it points by negative offset to the current location.
 *	The current value of the attribute is then set to the offset
 *	if the production rhs table of the current attribute.
 */

fixrhsinhattr (lattr, rattr, strng)

ATTRIB	*lattr;
ATTRIB	*rattr;
char	*strng;

{
	register	int	offset, value;

	mkrhstoken (0);
	offset = rhsoffset ();

	/*
	 *	If there is not previous reference to this
	 *	attribute in the rhs attribute table, define
	 *	it.  A value will be provided later.
	 */

	if (!defattr (rattr, strng)) {
		addattr (rattr, strng);

		/*
		 *	Link the rhs inherited attribute to the
		 *	lhs inherited attribute of the same name
		 *	if there is one.
		 */

		if (defattr (lattr, strng)) {
			if ((value = getattr (lattr, strng)) <= 0) {
				fprintf (stderr, "%d: %s %s.\n", line,
				"Undefined attribute --", strng);
				seterror ();
			}

			/*
			 *	Adjust value so that it is a rhs token
			 *	instead of an lhs inherited attribute
			 *	number for use below.
			 */

			value -= 1;
		} else {
			fprintf (stderr, "%d: %s %s.\n", line,
			"Undefined attribute", strng);
			seterror ();
			value = offset;
		}
	} else {
		/*
		 *	Set the previous reference to this attribute
		 *	to point by negative offset to the location
		 *	of the current attribute in the production rhs.
		 */

		value = getattr (rattr, strng);
		setrhstoken (value, value - offset);
	}

	if (value >= 0) {
		setrhstoken (value, value - offset);
	}

	/*
	 *	Set the value of the rhs inherited attribute to point
	 *	to the current (most recent) inheritance reference.
	 */

	setattr (rattr, strng, offset);
}

/*
 *	This routine is used to process rhs synthesized attributes
 *	as they are encountered during the parse of the grammar.
 */

fixrhssynattr (lattr, rattr, strng)

ATTRIB	*lattr;
ATTRIB	*rattr;
char	*strng;

{
	register	int	offset;

	mkrhstoken (0);
	offset = rhsoffset ();

	/*
	 *	Check to see whether this attribute has had a value
	 *	defined for it in the rhs table, or whether it has
	 *	the same name as an inherited lhs attribute.  Either
	 *	case represents redefination of the attribute and is
	 *	an error.
	 */

	if (defattr (rattr, strng)
	|| (defattr (lattr, strng) && getattr (lattr, strng) > 0)) {
		fprintf (stderr, "%d: %s %s.\n", line,
		"Redefined attribute --", strng);
		seterror ();
	} else {
		addattr (rattr, strng);
	}

	/*
	 *	The attribute reference is OK, so just make it point
	 *	to the offset of the current symbol in the production rhs.
	 */

	setattr (rattr, strng, offset);
}
