/*+
 *	Module:		llfmq.c
 *
 *	Purpose:	Provides subroutines which calculate the least
 *			cost terminal derivable strings for each non-
 *			terminal grammar symbol as well as the error
 *			recovery strings for each pair of non-terminal
 *			and terminal symbols in the grammar.  These
 *			strings will be used to generate the FMQ error
 *			recovery tables as described in Fischer, Milton,
 *			Quiring, "Efficient LL(1) Error Correction and
 *			Recovery Using Only Insertions", Acta Informatica 13,
 *			pp 141 - 154, 1980.
 *
 *	Invocation:	mkFMQstrngs ();
 *
 *			prFMQstrngs ();
 *
 *	Parameters:
 *
 *	  Input:	None.
 *
 *	  Output:	The error recovery string tables are generated
 *			in XXX.
 *
 *	Author:		D. Bocking
 *
 *	Date:		4 December, 1984
-*/

#include	"llsymtab.h"
#include	"llparam.h"
#include	"llcosts.h"
#include	"llinstr.h"
#include	<stdio.h>
#include	"llglobal.h"

/*
 *	The arrays of error recovery string pointers.
 */

INSTR	*S [MAXSIZE];
INSTR	*E [MAXTABLE];

/*
 *	The following routine calculates the minimum cost terminal
 *	derivable strings for each non-terminal in the grammar.
 */

mkFMQstrngs ()

{
	register	int	i, j;
	int	intarr [1], nochange, prodcost (), errcost();
	INSTR	*lambda, *mknullintstr (), *defnintstr ();
#ifdef	DEBUG
	struct	symtab	*getsymaddr ();
#endif

	/*
	 *	Set all terminal symbol costs to be 1 and
	 *	all nonterminal symbol costs to be infinite.
	 */

	for (i = 0; i < fnt; i++) {
		setcost (i, 1);

		intarr [0] = i;
		S [i] = defnintstr (intarr, 1);
	}

	for ( ; i < fa; i++) {
		setcost (i, INFINITY);

		S [i] = UNDEFINED;
	}

	/*
	 *	Calculate the least cost terminal derivable strings for
	 *	each of the non-terminal grammar symbols.  This is 
	 *	Algorithm A.1 from the paper.  Most of the work is done
	 *	by prodcost below.
	 */

	do {
		nochange = 1;

		foreachprod (prodcost, &nochange);

	} while (nochange == 0);

#ifdef	DEBUG
	for (i = 0; i < fa; i++) {
		fprintf (stderr, "\t%s\t%d\n", getsymaddr (i) -> sname,
		getcost (i));
	}

	for (i = fnt; i < fa; i++) {
		fprintf (stderr, "\n%s\n", getsymaddr (i) -> sname);
		dumpintstr (S [i]);
	}
#endif

	/*
	 *	Calculate the error correction strings for
	 *	the E table.  This is Algorithm A.2 from the
	 *	the paper.  Most of the work is done by errcost
	 *	below.
	 */

	lambda = mknullintstr ();
	for (i = 0; i < fa; i++) {
		for (j = 0; j < fnt; j++) {
			if (i == j) {
				E [i * fnt + j] = lambda;
			} else {
				E [i * fnt + j] = UNDEFINED;
			}
		}
	}

	do {
		nochange = 1;

		foreachprod (errcost, &nochange);

	} while (nochange == 0);
}

/*
 *	The following routine prints the error corrections tables
 *	out into the tables produced by the parser generator.
 */

prFMQstrngs ()

{
	register	int	i, j, length, offset;
	register	char	*symname;
	register	INSTR	*strng;
	struct	symtab	*getsymaddr ();

	fprintf (tables, "\n/* FMQ Error Recovery Tables */\n\n");

	fprintf (tables, "/* Symbol Insertion Cost Table */\n\n");
	fprintf (tables, "int llfmqcosts [] = {\n");
	for (i = 0; i < fa; i++) {
		fprintf (tables, "\t%d,\t/* %s */\n", getcost (i),
		getsymaddr (i) -> sname);
	}
	fprintf (tables, "};\n");

	fprintf (tables, "\n/* Error String Token Table */\n\n");
	fprintf (tables, "short llfmqstrngs [] = {\n");
	fprintf (tables, "\n/* The Null String */\n\t-1,\n");
	for (i = 0; i < fa; i++) {
		if ((strng = S [i]) != UNDEFINED && lenintstr (strng) != 0) {
			fprintf (tables, "\n/* S [%s] */\n",
			getsymaddr (i) -> sname);
			printstr (strng, 20);
		}
	}

	for (i = fnt; i < fa; i++) {
		symname = getsymaddr (i) -> sname;
		for (j = 0; j < fnt; j++) {
			if ((strng = E [i * fnt + j]) != UNDEFINED
			&& lenintstr (strng) != 0) {
				fprintf (tables, "\n/* E [%s, %s] */\n",
				symname, getsymaddr (j) -> sname);
				printstr (strng, 20);
			}
		}
	}
	fprintf (tables, "};\n");

	offset = 1;

	fprintf (tables, "\n/* Least Cost Terminal Strings */\n\n");
	fprintf (tables, "short *llfmqS [] = {\n");
	for (i = 0; i < fa; i++) {
		if ((strng = S [i]) == UNDEFINED) {
			fprintf (tables, "\t(short *) (-1),\n");
		} else if ((length = lenintstr (strng)) == 0) {
			fprintf (tables, "\t&llfmqstrngs [0],\n");
		} else {
			fprintf (tables, "\t&llfmqstrngs [%d],\n", offset);
			offset += length + 1;
		}
	}
	fprintf (tables, "};\n");

	fprintf (tables, "\n/* Error Correction Strings */\n\n");
	fprintf (tables, "short *llfmqE [] = {\n");
	for (i = fnt; i < fa; i++) {
		for (j = 0; j < fnt; j++) {
			if ((strng = E [i * fnt + j]) == UNDEFINED) {
				fprintf (tables, "\t(short *) (-1),\n");
			} else if ((length = lenintstr (strng)) == 0) {
				fprintf (tables, "\t&llfmqstrngs [0],\n");
			} else {
				fprintf (tables, "\t&llfmqstrngs [%d],\n",
				offset);
				offset += length + 1;
			}
		}
	}
	fprintf (tables, "};\n");
}

/*
 *	The following routine calculates the cost of a production
 *	right hand side and determines whether or not this is
 *	the smallest cost rhs available for its left hand side
 *	non-terminal.
 */

prodcost (n, lhs, rhs, rhslen, nochange)

int	n;
struct	symtab	*lhs;
int	rhs [];
int	rhslen;
int	*nochange;

{
	register	int	i, newrhslen, rhscost;
	int	newrhs [200];
	INSTR	*setnintstr ();

	newrhslen = striprhs (lhs, rhs, rhslen, newrhs);
	i = lhs -> snum;

	rhscost = strncost (newrhs, newrhslen, S);
	if (rhscost < getcost (i)) {
#ifdef	DEBUG
		fprintf (stderr, "Prodcost sets new lhs cost:\n");
#endif
		setcost (i, rhscost);
		S [i] = setnintstr (newrhs, newrhslen, S);
#ifdef	DEBUG
		dumpintstr (S [i]);
#endif
		*nochange = 0;
	}
}

/*
 *	The following routine calculates the least cost error
 *	correction strings for each ordered pair of non-terminal
 *	and terminal symbols in the grammar.
 */

errcost (n, lhs, rhs, rhslen, nochange)

int	n;
struct	symtab	*lhs;
int	rhs [];
int	rhslen;
int	*nochange;

{
	register	int	i, j, k, k1, newrhslen;
	int	lhstoken, lhscost;
	int	mincost, mincostj, newrhs [200];
	INSTR	*Estr, **Eptr, *mincoststr, *mknullintstr (), *catnintstr ();
#ifdef	DEBUG
	struct	symtab	*getsymaddr ();
#endif

	newrhslen = striprhs (lhs, rhs, rhslen, newrhs);
	lhstoken = lhs -> snum;

#ifdef	DEBUG
	fprintf (stderr, "Errcost - production %d", n);
	fprintf (stderr, "  %s:\t", getsymaddr (lhstoken) -> sname);

	for (i = 0; i < newrhslen; i++) {
		fprintf (stderr, "%s ", getsymaddr (newrhs [i]) -> sname);
		if (i % 10 == 9) {
			fprintf (stderr, "\n\t\t");
		}
	}
	fprintf (stderr, "\n");
#endif

	for (i = 0; i < fnt; i++) {
		Eptr = &E [lhstoken * fnt + i];
		lhscost = strcost (*Eptr);
#ifdef	DEBUG
		fprintf (stderr, "E [%s, %s]:",
		getsymaddr (lhstoken) -> sname, getsymaddr (i) -> sname);
		fprintf (stderr, "  Cost: %d\n", lhscost);
		dumpintstr (*Eptr);
#endif

		/*
		 *	Find k and j such that
		 *
		 *	k = min ( C (X(1),...,X(j-1)) + C (E (X(j),i)
		 */

		mincost = lhscost;
		if (newrhslen != 0) {
			for (j = newrhslen - 1; j >= 0; j--) {
				Estr = E [newrhs [j] * fnt + i];
				k = strcost (Estr);

				if (j > 0 && k < INFINITY) {
					k1 = strncost (newrhs, j);
					if (k1 < INFINITY) {
						k += k1;
					} else {
						k = INFINITY;
					}
				}

				if (k < mincost) {
					mincost = k;
					mincostj = j;
					mincoststr = Estr;
				}
			}

			if (mincost < lhscost) {
				/*
				 *	E [lhstoken, i] = S (X(1),...,X(j-1))
				 *	cat E [X(j), i]
				 */
#ifdef DEBUG
				fprintf (stderr,
				"\tNew E [%s, %s] J: %d Cost: %d\n",
				getsymaddr (lhstoken) -> sname,
				getsymaddr (i) -> sname, mincostj, mincost);
#endif
				*Eptr = catnintstr (newrhs, mincostj,
				S, mincoststr);
#ifdef	DEBUG
				dumpintstr (*Eptr);
#endif
				*nochange = 0;
			}
		}
	}
}
