/*+
 *	Module:		llinstr.c
 *
 *	Purpose:	Provides subroutines to implement operations
 *			on intstrings.
 *
 *	Invocation:	instr = mknullintstr ();
 *
 *			instr = defnintstr (intarr, len);
 *
 *			instr = setnintstr (intarr, len, strtab);
 *
 *			instr = catnintstr (intarr, len, strtab, intstr);
 *
 *			printstr (intstr, mod);
 *
 *	Parameters:
 *
 *	  Input:	intarr - The address of the first element in an
 *			array of integer tokens which are to be made into an
 *			intstring using the strtab provided.
 *
 *			len - The length of the array of integers which
 *			is to become an intstring.
 *
 *			strtab - The address of a table containing the
 *			replacement strings for the tokens in intarr.
 *
 *			intstr - The address of an intstring which is to
 *			be concatenated onto a newly created inststring.
 *
 *			mod - The number of string entries to print on
 *			one line.
 *
 *	  Output:	instr - The address of the newly created intstring.
 *
 *	Author:		D. Bocking
 *
 *	Date:		5 December, 1984
-*/

#ifdef	DEBUG
#include	"llsymtab.h"
#endif
#include	"llinstr.h"
#include	<stdio.h>
#include	"llglobal.h"

/*
 *	Storage for the intstrings and a pointer to the next available
 *	entry.
 */

#define	MAXINTSTR	60000

INSTR	llintspace [MAXINTSTR];
int	llnextint	=	0;

/*
 *	The following routine creates a null intstring.
 */

INSTR	*
mknullintstr ()

{
	if (llnextint >= MAXINTSTR) {
		fprintf (stderr, "Out of IntString space!\n");
		exit (1);
	}

	llintspace [llnextint] = -1;
	return (&llintspace [llnextint++]);
}

/*
 *	The following routine defines an intstring which contains
 *	the tokens passed in the argument array.
 */

INSTR	*
defnintstr (intarr, len)

int	intarr [];
int	len;

{
	register	INSTR	*base;
	register	int	i;

	if (llnextint + len + 1 >= MAXINTSTR) {
		fprintf (stderr, "Out of IntString Space!\n");
		exit (1);
	}

	base = &llintspace [llnextint];
	for (i = 0; i < len; i++) {
		llintspace [llnextint++] = intarr [i];
	}
	llintspace [llnextint++] = -1;

	return (base);
}

/*
 *	The following routine creates a new intstring from an
 *	array of integers of known length.  Each of the integers
 *	is a grammar token which is to be replaced by its corresponding
 *	least cost terminal derivable string in the newly created
 *	intstring.
 */

INSTR	*
setnintstr (intarr, len, strtab)

int	intarr [];
int	len;
INSTR	*strtab [];

{
	register	int	length;
	register	INSTR	*base;

	length = lentokenstr (intarr, len, strtab);
	if (llnextint + length + 1 > MAXINTSTR) {
		fprintf (stderr, "Out of IntString space!\n");
		exit (1);
	}

	base = &llintspace [llnextint];
	cpytokenstr (intarr, len, strtab, base);
	llnextint += length;
	llintspace [llnextint++] = -1;

	return (base);
}

/*
 *	The following routine creates a new intstring from an array
 *	of integers of known length, as does setnintstr, and then
 *	concatenates an already defined intstring to it.
 */

INSTR	*
catnintstr (intarr, len, strtab, intstr)

int	intarr [];
int	len;
INSTR	*strtab [];
INSTR	*intstr;

{
	register	int	len1, len2;
	register	INSTR	*base;
	INSTR	*setnintstr ();

	len1 = lentokenstr (intarr, len, strtab);
	len2 = lenintstr (intstr);

	if (llnextint + len1 + len2 + 1 > MAXINTSTR) {
		fprintf (stderr, "Out of IntString space!\n");
		exit (1);
	}

	base = setnintstr (intarr, len, strtab);

	llnextint--;	/* Back up over the appended null integer */
	cpyintstr (intstr, &llintspace [llnextint]);
	llnextint += len2;

	llintspace [llnextint++] = -1;
	return (base);
}

/*
 *	The following subroutine prints the contents of a string
 *	in the output tables, mod entries to a line.
 */

printstr (intstr, mod)

INSTR	*intstr;
int	mod;

{
	register	INSTR	*sptr;
	register	int	count;

	count = 0;
	sptr = intstr;

	fprintf (tables, " \t");
	do {
		fprintf (tables, "%d,", *sptr);
		if (count++ % mod == mod - 1) {
			fprintf (tables, "\n\t");
		}
	} while (*sptr++ != -1);
	fprintf (tables, "\n");
}

/*
 *	The following subroutine calculates the length of a given intstring.
 */

lenintstr (intstr)

INSTR	*intstr;

{
	register	int	len;
	register	INSTR	*insptr;

	insptr = intstr;
	len = 0;

	if (insptr != UNDEFINED) {
		while (*insptr++ != -1) {
			len++;
		}
	}

	return (len);
}

/*
 *	The following routine calculates the length of the replacement
 *	strings corresponding to a list of tokens.
 */

lentokenstr (intarr, len, strtab)

int	intarr [];
int	len;
INSTR	*strtab [];

{
	register	int	i, totallen;

	totallen = 0;
	for (i = 0; i < len; i++) {
		totallen += lenintstr (strtab [intarr [i]]);
	}

	return (totallen);
}

/*
 *	The following subroutine copies an intstring from one
 *	place to another, stopping after the terminating null
 *	integer has been copied.
 */

cpyintstr (from, to)

INSTR	*from;
INSTR	*to;

{
	register	INSTR	*fptr, *tptr;

	fptr = from;
	tptr = to;

	if (fptr != UNDEFINED) {
		do {
			*to++ = *from;
		} while (*from++ != -1);
	} else {
		*to++ = -1;
	}
}

/*
 *	The following subroutine copies the replacement strings
 *	of a list of input tokens to a designated address.
 */

cpytokenstr (intarr, len, strtab, whereto)

int	intarr [];
int	len;
INSTR	*strtab [];
INSTR	*whereto;

{
	register	int	i;
	register	INSTR	*whereptr, *fromptr;

	whereptr = whereto;

	for (i = 0; i < len; i++) {
		fromptr = strtab [intarr [i]];
		cpyintstr (fromptr, whereptr);
		whereptr += lenintstr (fromptr);
	}
}

#ifdef	DEBUG

/*
 *	The following routine is used to dump out the printable
 *	names of the tokens in an intstring during debugging.
 */

dumpintstr (strng)

INSTR	*strng;

{
	register 	INSTR	*sptr;
	struct	symtab	*getsymaddr ();

	if (strng == UNDEFINED) {
		fprintf (stderr, "\tUNDEFINED\n");
	} else {
		for (sptr = strng; *sptr != -1; sptr++) {
			fprintf (stderr, "\t%s\n", getsymaddr (*sptr) -> sname);
		}
	}
}
#endif
