/*+
 *	Module:		llmain.c
 *
 *	Purpose:	Provides the main grammar processing of the
 *			table generator.
 *
 *	Invocation:	llgen file (shell command line)
 *
 *	Parameters:
 *
 *	  Input:	file - The name of the file containing the
 *			specification of the grammar and supporting
 *			subroutines.
 *
 *	  Output:	"ll.tab.c" - The tables generated are placed
 *			in this file in the current directory.
 *
 *	Author:		D. Bocking
 *
 *	Date:		22 November, 1984
-*/

#include	"llscan.h"
#include	"llsymtab.h"
#include	"llprod.h"
#include	<stdio.h>
#include	"llparam.h"
#include	"llglobal.h"

/*
 *	Storage for the bit matrices produced by the grammar
 *	analyser.
 */

char	F [LENGTH],		/* for F and F_star relations. */
	B [LENGTH],		/* for B relation and 'follow' relation. */
	L [LENGTH],		/* for L and L_star relations. */
	W [LENGTH];		/* 'work' relation. */

/* alternate matrix names are as follows: */

#define	Fstar	F
#define	Lstar	L
#define	follow	B

char	nullable [(MAXSIZE+7)>>3];

/* the internal copy of the final parse table: */

int	M [MAXTABLE];

main(argc, argv)

int argc;
char *argv[];

{
	register	int	moretocopy, bytesread;
	char	buffer [512];

	/*
	 *	Check for proper invocation of the generator.
	 */

	if(argc != 2){
		printf("usage:  llgen FILE\n");
		exit (1);
	}

	/*
	 *	Try to open the grammar specification file.
	 */

	if((input = fopen(argv[1],"r")) == NULL){
		printf("llgen cannot open '%s'\n",argv[1]);
		exit (1);
	}

	/*
	 *	Create the tables output file ... "ll.tab.c"
	 */

	if ((tables = fopen ("ll.tab.c", "w")) == NULL) {
		fprintf (stderr, "Llgen cannot create ll.tab.c\n");
		exit (1);
	}

#ifdef VERBOSE
	debug = fopen("ll.debug","w");
	if(debug < 0){
		printf("llgen cannot create ll.debug\n");
		exit (1);
	}
#endif

	/*
	 *	Read the grammar.
	 */

	moretocopy = read_grammar();

	/*
	 *	Generate the parser tables.
	 */

	gen_tables ();

	/*
	 *	Copy any remaining subroutines, etc. to the output.
	 */

	if (moretocopy) {
		while ((bytesread = fread (buffer, 1, 512, input)) > 0) {
			fwrite (buffer, 1, bytesread, tables);
		}
	}

	exit(0);
}


read_grammar ()

{
	register	int	token;

	line = 1;

	token = rdgrammar ();
	if (token == PERCENT || token == TKEOF) {
		if (wasanerror () == 0) {
			prgrammar ();
#ifdef	DEBUG
			fflush (tables);
#endif
		} else {
			fprintf (stderr, "%s %s.\n",
			"Faulty translator specification --",
			"No tables generated");
			exit (1);
		}
	} else {
		fprintf (stderr, "%d: Specification syntax error!\n",
		line);
		exit (1);
	}

	return (token == PERCENT ? 1 : 0);
}

gen_tables ()

{
	int	change, chknullable (), mkFBL (), mktable ();

	/*
	 *	Find nullable non-terminal symbols.
	 */

	do {
		change = 0;
		foreachprod (chknullable, &change);
	} while(change);

	/*
	 *	Calculate the F, B, and L relations.
	 */

	foreachprod (mkFBL, 0);

	/*
	 *	Augment the grammar with the starting production.
	 */

	set (B, fa, startsymbol -> snum, endsymbol -> snum);

#ifdef VERBOSE
	fprintf(debug,"\n\nF relation:\n\n");
	put_relation(F, fa);
	fprintf(debug,"\n\nB relation:\n\n");
	put_relation(B, fa);
	fprintf(debug,"\n\nL relation\:\n\n");
	put_relation(L, fa);
#endif

	/*
	 *	Calculate the Fstar relation.
	 */

	warshall(F,fa,0);
	left_recursion();

#ifdef VERBOSE
	fprintf(debug,"\n\nF-star relation:\n\n");
	put_relation(Fstar, fa);
#endif

	/*
	 *	Calculate Lplus relation.
	 */

	warshall(L, fa, 1);

#ifdef VERBOSE
	fprintf(debug,"\n\nL-star relation:\n\n");
	put_relation(Lstar, fa);
#endif

	/*
	 *	Calculate the FOLLOW relation.
	 */

	multiply (fa, Lstar, B, W);
	multiply (fa, W, Fstar, follow);

#ifdef VERBOSE
	fprintf(debug,"\n\n'follow' relation (L* B F*)\n\n");
	put_relation(follow, fa);
#endif

	/*
	 *	Build the parse table.
	 */

	foreachprod (mktable, 0);

	/*
	 *	Finish generating the tables.
	 */

	finish();
#ifdef	FMQ
	mkFMQstrngs ();
	prFMQstrngs ();
#endif
}

/*
 *	The following subroutine is called from gen_tables by
 *	foreachprod to check the right hand side of a production
 *	to determine whether the left hand side non terminal is
 *	a nullable non-terminal symbol.
 */

chknullable (n, lhs, rhs, rhslen, change)

int	n;
struct	symtab	*lhs;
int	rhs [];
int	rhslen;
int	*change;

{
	int	empty, i, newrhslen, newrhs [200];

	newrhslen = striprhs (lhs, rhs, rhslen, newrhs);

	empty = 1;
	for (i = 0; i < newrhslen; i++) {
		if (test (nullable, 0, 0, newrhs [i]) == 0) {
			empty = 0;
		}
	}

	if (empty && test (nullable, 0, 0, lhs -> snum) == 0){
		set (nullable, 0, 0, lhs -> snum);
		*change = 1;
	}
}

/*
 *	The following routine is called by gen_tables through
 *	foreachprod to extract the F, B, and L bit matrices
 *	from the grammar tables.
 */

mkFBL (n, lhs, rhs, rhslen, p)

int	n;
struct symtab	*lhs;
int	rhs [];
int	rhslen;
int	p;

{
	register	int	i, j, k, l, lhstoken, newrhslen;
	int	newrhs [200];

	newrhslen = striprhs (lhs, rhs, rhslen, newrhs);

	lhstoken = lhs -> snum;

	/*
	 *	First, calculate the F relation.
	 */

	for (i = 0; i < newrhslen; i++) {
		j = newrhs [i];
		set (F, fa, lhstoken, j);
		if (!test (nullable, 0, 0, j))
			break;
	}

	/*
	 *	Calculate the B relation.
	 */

	for (i = 0; i < newrhslen; i++) {
		j = newrhs [i];
		for (k = i + 1; k < newrhslen; k++) {
			l = newrhs [k];
			set (B, fa, j, l);
			if (!test (nullable, 0, 0, l))
				break;
		}
	}

	/*
	 *	Finally, calculate the L relation.
	 */

	for (i = newrhslen - 1; i >= 0; i--) {
		j = newrhs [i];
		set (L, fa, j, lhstoken);
		if (!test (nullable, 0, 0, j))
			break;
	}
}

left_recursion()

{
	register	int	i;
	struct	symtab	*getsymaddr ();

	/* go thru F-plus relation, check for left-recursive */
	/* nonterminals, and set diagonal elements to get */
	/* F-star relation. */

	for (i = 0; i < fa ; i++) {
		if (i >= fnt && test (F, fa, i, i)) {
			fprintf (stderr,
			"Nonterminal %s is left recursive\n",
			getsymaddr (i) -> sname);

#ifdef VERBOSE
			fprintf (debug,
			"Nonterminal %s is left recursive\n",
			getsymaddr (i) -> sname);
#endif
		}
		set (F, fa, i, i);
	}
}

/*
 *	The following routine is called by gen_tables through
 *	foreachprod to make entries in the final parse table.
 */

mktable (n, lhs, rhs, rhslen, p)

int	n;
struct	symtab	*lhs;
int	rhs [];
int	rhslen;
int	p;

{
	register	int	i, j, k, newrhslen;
	int	newrhs [200];

	newrhslen = striprhs (lhs, rhs, rhslen, newrhs);

	for (i = 0; i < newrhslen; i++) {
		j = newrhs [i];

		if (j < fnt) {
			setM (lhs -> snum, j, n + 1);
		} else {
			for (k = 0; k < fnt; k++) {
				if (test (Fstar, fa, j, k)) {
					setM (lhs -> snum, k, n + 1);
				}
			}
		}

		if (!test (nullable, 0, 0, j)) {
			return;
		}
	}

	for (k = 0; k < fnt; k++) {
		if (test (follow, fa, lhs -> snum, k)) {
			setM (lhs -> snum, k, n + 1);
		}
	}
}

setM (nt, t, i)

int nt, t, i;

{
	register	int	*mp;
	register	int	v;
	struct	symtab	*getsymaddr ();

	mp = M + (nt - fnt)*fnt + t;
	v = (*mp);
	if(v > 0 && v != i){
		fprintf (stderr,
		"Conflict: entry [%s, %s] productions %d and %d\n",
		getsymaddr (nt) -> sname, getsymaddr (t) -> sname, v, i);
#ifdef VERBOSE
		fprintf (debug,
		"Conflict: entry [%s, %s] productions %d and %d\n",
		getsymaddr (nt) -> sname, getsymaddr (t) -> sname, v, i);
#endif
	}
	else
		*mp = i;
}

/*
 *	The following routine produces the remainder of the parse tables.
 */

finish ()

{
	register	int	nt, t;
	register	int	i, j, prod;
	struct	production	*pp;
	struct	symtab	*getsymaddr ();

#ifdef VERBOSE
	fprintf(debug,
	"\n\nThe significant region of the parse table is:\n\n     ");

	for(t = 0; t < fnt ; t++)
		fprintf(debug, "____%3d", t);

	for(nt = fnt ; nt < fa ; nt++){
		fprintf(debug,"\n%3d |", nt);
		for(t = 0; t < fnt ; t++){
			prod = M[(nt - fnt)*fnt + t];
			if(prod)
				fprintf(debug," %3d", prod);
			else
				fprintf(debug,"   .");
		}
	}
	fprintf(debug,"\n\nSymbol table:\n\n");

	for(t = 0 ; t < fnt ; t++)
		fprintf(debug,"%3d  %d %s\n",t, test(nullable,0,0,t)
		,symbol[t]);

	fprintf(debug,"--------------\n");
	for(nt = fnt; nt < fa; nt++)
		fprintf(debug,"%3d  %d %s\n",nt,test(nullable,0,0,nt)
		,symbol[nt]);
#endif

	/*
	 *	Print out the parse table.
	 */

	fprintf (tables, "\n/* The Parser Action Table */\n\n");
	fprintf (tables, "int llpt[] = {");
	prod = 0;

	for (i = fnt; i < fa; i++) {
		/*
		 *	One row for each non terminal symbol in
		 *	the grammar.
		 */

		fprintf (tables, "\n\n/* %s */\n\t", getsymaddr (i) -> sname);

		for (j = 0; j < fnt; j++) {
			if (j % 16 == 15) {
				fprintf (tables, "\n\t");
			}

			fprintf (tables, "%d, ", M [prod++]);
		}
	}
	fprintf(tables, "\n};\n");

}

/* Routine to print out a relation. */

put_relation(ar, arl)

char *ar;
int arl;

{
	register char *r;
	register int i,j;

	r = ar;
	for(i = 0; i <  arl ; i++){
		for(j = 0 ; j < arl ; j++)
			fputc('0' + test(r,arl,i,j),debug);
		fputc('\n',debug);
	}
}

/*
 *	The following routine is used to strip the attributes and
 *	action symbols out of the right hand side of a production
 *	to simplify the job of analyzing the grammar.
 */

striprhs (lhs, rhs, rhslen, newrhs)

struct	symtab	*lhs;
int	rhs [];
int	rhslen;
int	newrhs [];

{
	register	int	i, j, k;
	register struct	symtab	*sp;

#ifdef	DEBUG
	fprintf (stderr, "Striprhs - lhs %s rhslen %d.\n",
	lhs -> sname, rhslen);
#endif

	i = lhs -> inherit + lhs -> synth;

	k = 0;
	while (i < rhslen) {
		sp = (struct symtab *) rhs [i++];
#ifdef	DEBUG
		fprintf (stderr, "\t%s\n", sp -> sname);
#endif
		if ((j = sp -> snum) < fa) {
			newrhs [k++] = j;
		}
		i += sp -> inherit + sp -> synth;
	}

	return (k);
}
