
/*+
 *	Module:		ll1.c
 *
 *	Purpose:	This module provides bit matrix operations test,
 *			set, and warshall for use in the ll parser generator.
 *
 *	Invocation:	bitonoroff = test (matrix, rowlen, row, col);
 *
 *			set (matrix, rowlen, row, col);
 *
 *			warshall (matrix, rowlen, reflexive);
 *
 *			multiply (rowlen, matrixA, matrixB, matrixC);
 *
 *	Parameters:
 *
 *	  Input:	matrix - The address of the first character of
 *			the storage containing the bit matrix.
 *
 *			matrixA, matrixB, matrixC - Multiply returns
 *			the matrix product matrixA * matrixB in matrixC.
 *
 *			rowlen - The length in bits of each row of the
 *			bit matrix.
 *
 *			row - The number of the row containing the bit
 *			to be tested or set.
 *
 *			col - The number of the column containing the bit
 *			to be tested or set.
 *
 *			reflexive - A flag indicating whether or not the
 *			reflexive transitive closure as opposed to the
 *			transitive closure of the bit matrix is to be
 *			generated.
 *
 *	  Output:	bitonoroff - Reflects the status of the tested bit.
 *
 *	Author:		D. Bocking
 *
 *	Date:		5 March, 1982
-*/

/*+
 *	MODIFICATION	1
 *
 *	Purpose:	Rewritten in C for portability.
 *
 *	Author:		D. Bocking
 *
 *	Date:		5 March, 1982
-*/

#include	"llparam.h"

test (matrix, rowlen, row, col)

char	*matrix;
int	rowlen, row, col;

{
	register int	rowbytes, byteindex, colbytes, colbits;

	rowbytes = (rowlen + 7) / 8;

	colbytes = col / 8;
	colbits = col % 8;

	byteindex = row * rowbytes + colbytes;

	return ((matrix [byteindex] & (1 << colbits)) ? 1 : 0);
}

set (matrix, rowlen, row, col)

char	*matrix;
int	rowlen, row, col;

{
	register int	rowbytes, byteindex, colbytes, colbits;

	rowbytes = (rowlen + 7) / 8;

	colbytes = col / 8;
	colbits = col % 8;

	byteindex = row * rowbytes + colbytes;

	matrix [byteindex] |= (1 << colbits);
}

warshall (matrix, rowlen, reflexive)

char	*matrix;
int	rowlen;
int	reflexive;

{
	int	i, j, k;
	register	char	*rowi, *rowk;
	register	int	rowbytes;

	rowbytes = (rowlen + 7) / 8;
	for (k = 0; k < rowlen; k++) {
		for (i = 0; i < rowlen; i++) {
			if (test (matrix, rowlen, i, k)) {
				rowi = &matrix [i * rowbytes];
				rowk = &matrix [k * rowbytes];
				for (j = 0; j < rowbytes; j++) {
					*rowi++ |= *rowk++;
				}
			}
		}
	}

	if (reflexive) {
		for (i = 0; i < rowlen; i++) {
			set (matrix, rowlen, i, i);
		}
	}
}

/*
 *	Multiply two bit matrices together.  This is done
 *	by transposing columns 8 at a time so that they are
 *	in column major format (see transcol below) and then
 *	anding the rows and columns together a byte at a time.
 */

char	columns [(MAXSIZE + 8) * 8];	/* Storage for 8 transposed columns */

multiply (rowlen, matrixA, matrixB, matrixC)

int	rowlen;
char	*matrixA, *matrixB, *matrixC;

{
	register	char	*rp, *cp;
	int	rowbytes, i, j, k, l, matbytes, rows, nrows;

	rows = rowlen % 8;
	if (rows == 0) {
		rows = 8;
	}

	rowbytes = (rowlen + 7) / 8;
	matbytes = rowbytes * rowlen;

	/*
	 *	Clear the result matrix to begin with.
	 */

	for (rp = &matrixC [matbytes - 1]; rp >= matrixC; rp--) {
		*rp = 0;
	}

	/*
	 *	For each group of (up to) 8 columns ...
	 */

	for (k = 0; k < rowbytes; k++) {

		/*
		 *	Transpose the group of columns.
		 */

		transcol (rowbytes, rows, &matrixB [k], columns);

		nrows = 8;
		if (k == rowbytes - 1) {
			nrows = rows;
		}

		/*
		 *	For each of the transposed columns ...
		 */

		for (j = 0; j < nrows; j++) {

			/*
			 *	Multiply by each of the rows of the first
			 *	bit matrix.
			 */

			for (i = 0; i < rowlen; i++) {
				cp = &columns [j * rowbytes];
				rp = &matrixA [i * rowbytes];

				for (l = 0; l < rowbytes; l++) {
					if (*rp++ & *cp++) {
						set (matrixC, rowlen, i,
						k * 8 + j);
						break;
					}
				}
			}
		}
	}
}

/*
 *	This routine is called by the multiply routine to transpose
 *	a group of up to 8 columns of a bit matrix so that the columns
 *	are stored in column major form, as opposed to the row major
 *	form used elsewhere throughout the parser generator.  This is
 *	done so that multiplication of the rows of one bit matrix
 *	by the columns of another can be done more efficiently, a byte
 *	at a time, instead of a bit at a time.  (Large matrices take
 *	a LONG time!).
 */

transcol (rowbytes, lastbits, rowptr, colptr)

int	rowbytes;	/* Number of bytes in a row of the matrix */
int	lastbits;	/* Number of bits in the last byte */
char	*rowptr;	/* Ptr to first row of column to be done */
char	*colptr;	/* Ptr to first column of output */

{
	char	rowarr [8];
	register	char	bitmask, outbyte, *cp;
	int	i, j, k, rows;

	for (k = 0; k < rowbytes; k++) {
		rows = 8;
		if (k == rowbytes - 1) {
			rows = lastbits;
		}

		for (i = 0; i < rows; i++) {
			rowarr [i] = *rowptr;
			rowptr += rowbytes;
		}

		for (i = rows; i < 8; i++) {
			rowarr [i] = 0;
		}

		cp = colptr++;
		bitmask = 1;
		for (i = 0; i < 8; i++) {
			for (j = 7; j >= 0; j--) {
				outbyte <<= 1;
				if (rowarr [j] & bitmask) {
					outbyte |= 1;
				}
			}

			*cp = outbyte;
			cp += rowbytes;
			bitmask <<= 1;
		}
	}
}
