/*+
 *	Module:		llparse.c
 *
 *	Purpose:	Provides subroutines which parse the input
 *			grammar specification and build the internal
 *			tabular representation of the productions.
 *
 *	Invocation:	token = rdgrammar ();
 *
 *	Parameters:
 *
 *	  Input:	source - A standard IO package file descriptor
 *			from which the grammar specification will be
 *			read.  This is an extern global variable, used
 *			by llscan, called from these routines.
 *
 *	  Output:	token - The integer value of the token which
 *			resulted in the termination of the parse of the
 *			input grammar.  On normal termination, this should
 *			be either PERCENT or TKEOF.  On failure, token
 *			contains the value -1.
 *
 *	Author:		D. Bocking
 *
 *	Date:		15 October, 1984
-*/

#include	<stdio.h>
#include	"llscan.h"
#include	"llattrib.h"
#include	"llsymtab.h"

union	tokenval	tokenval;
struct	symtab	*lookup (), *startsymbol, *endsymbol;

rdgrammar ()

{
	register	int	token;
	int	*mkrhsbase ();

#ifdef	DEBUG
	fprintf (stderr, "Entering rdgrammar.\n");
#endif
	startsymbol = (struct symtab *) 0;
	endsymbol = lookup ("lleof");
	endsymbol -> stype = SYMRHS;

	do {
		token = rdproduction ();
	} while (token == SEMI);
	mkproduction ((struct symtab *) -1, mkrhsbase ());

	if (token != PERCENT && token != TKEOF) {
		token = -1;
	}

#ifdef	DEBUG
	fprintf (stderr, "Leaving rdgrammar.\n");
#endif
	return (token);
}

rdproduction ()

{
	register	int	i, token, *rhs;
	register	struct	symtab	*lhsym;
	int	*mkrhsbase ();
	ATTRIB	lattrib;

#ifdef	DEBUG
	fprintf (stderr, "Entering rdproduction.\n");
#endif
	token = llscan ();
	if (token == PERCENT || token == TKEOF) {
		return (token);
	}

	/* Have a new production ... look for its left hand side */

	if (token == IDENT) {
		lhsym = lookup (tokenval.tk_str);
		if (startsymbol == (struct symtab *) 0) {
			startsymbol = lhsym;
		}

		clrattr (&lattrib);
		rhs = mkrhsbase ();
		token = rdlhsattr (lhsym, &lattrib);

		if (token == COLON) {

			/* Process each right hand side */
			do {
				mkproduction (lhsym, rhs);
				token = rdrhs (&lattrib);
				if (token == BAR) {
					rhs = mkrhsbase ();
					i = lhsym -> inherit + lhsym -> synth;
					while (i--) {
						mkrhstoken (0);
					}
				}
			} while (token == BAR);
		}
	}

#ifdef	DEBUG
	fprintf (stderr, "Leaving rdproduction.\n");
#endif
	return (token);
}

rdrhs (lattrib)

ATTRIB	*lattrib;

{
	register	int	token, rhtype;
	register	struct	symtab	*rhsym;
	int	fixlhssynattr ();
	ATTRIB	rattrib;

#ifdef	DEBUG
	fprintf (stderr, "Entering rdrhs.\n");
#endif
	clrattr (&rattrib);
	token = llscan ();
	do {
		if (token == IDENT || token == ACTION) {
			rhsym = lookup (tokenval.tk_str);
			mkrhstoken ((int) rhsym);

			rhtype = (token == IDENT ? SYMRHS : SYMACT);

			token = rdrhsattr (lattrib, rhsym, &rattrib);

			if (rhsym -> stype == 0) {
				rhsym -> stype = rhtype;
			}
		}
	} while (token == IDENT || token == ACTION);

	/*
	 *	Fix up the lhs synthesized attributes i
	 *	in this production.
	 */

	foreachattr (lattrib, fixlhssynattr, &rattrib);
#ifdef	DEBUG
	fprintf (stderr, "Leaving rdrhs.\n");
#endif
	return (token);
}

rdlhsattr (sym, attr)

struct	symtab	*sym;
ATTRIB	*attr;

{
	register	int	token, i, s;

	/* Look for inherited attributes */

#ifdef	DEBUG
	fprintf (stderr, "Entering rdlhsattr.\n");
#endif
	i = 0;
	token = llscan ();

	if (token == LESSTHN) {
		do {
			token = llscan ();

			if (token == IDENT) {
				i++;
				fixlhsattr (attr, tokenval.tk_str, i);
				token = llscan ();
			} else {
				break;
			}
		} while (token == COMMA);
	}

	/* Look for synthesized attributes */

	s = 0;
	if (token == GRTRTHN) {
		do {
			token = llscan ();
			if (token == IDENT) {
				s++;
				fixlhsattr (attr, tokenval.tk_str, - (i + s));
				token = llscan ();
			} else {
				break;
			}
		} while (token == COMMA);
	}

	if (!chksymattr (sym, i, s)) {
		seterror ();
	}
	sym -> stype = SYMLHS;

#ifdef	DEBUG
	fprintf (stderr, "Leaving rdlhsattr.\n");
#endif
	return (token);
}

rdrhsattr (lattr, sym, rattr)

ATTRIB	*lattr;
struct	symtab	*sym;
ATTRIB	*rattr;

{
	register	int	token, i, s;

#ifdef	DEBUG
	fprintf (stderr, "Entering rdrhsattr.\n");
#endif
	/* Look for inherited attributes */

	i = 0;
	token = llscan ();
	if (token == LESSTHN) {
		do {
			token = llscan ();
			if (token == IDENT) {
				i++;
				fixrhsinhattr (lattr, rattr, tokenval.tk_str);
				token = llscan ();
			} else if (token == NUMBER) {
				i++;
				mkrhstoken (tokenval.tk_int);
				token = llscan ();
			} else {
				break;
			}
		} while (token == COMMA);
	}

	/* Look for synthesized attributes */

	s = 0;
	if (token == GRTRTHN) {
		do {
			token = llscan ();
			if (token == IDENT) {
				s++;
				fixrhssynattr (lattr, rattr, tokenval.tk_str);
				token = llscan ();
			} else {
				break;
			}
		} while (token == COMMA);
	}

	if (!chksymattr (sym, i, s)) {
		seterror ();
	}

#ifdef	DEBUG
	fprintf (stderr, "Leaving rdrhsattr.\n");
#endif
	return (token);
}
