/*+
 *	Module:		llrhs
 *
 *	Purpose:	Provides operations which manipulate the table of
 *			production right hand sides constructed during the
 *			parsing of the grammar specification.
 *
 *	Invocation:	rhsaddr = mkrhsbase ();
 *
 *			mkrhstoken (token);
 *
 *			offset = rhsoffset ();
 *
 *			token = getrhstoken (offset);
 *
 *			setrhstoken (offset, token);
 *
 *			prrhs ();
 *
 *			index = mkrhsindex (rhsaddr);
 *
 *	Parameters:
 *
 *	  Input:	rhsaddr - The address of the first token in the
 *			current right hand side.
 *
 *			token - An integer value to be stored as a token
 *			in a production right hand side.
 *
 *			offset - An integer offset into the right hand side
 *			of a production.
 *
 *			rhsaddr - The address of the base of a right hand
 *			side.
 *
 *	  Output:	offset - The offset in a production right hand side
 *			of the most recently added token.
 *
 *			token - The value of the token at the specified
 *			offset in the current production.
 *
 *			index - The index into the stored right hand side
 *			array of the specified right hand side.
 *
 *	Author:		D. Bocking
 *
 *	Date:		18 October, 1984
-*/

#include	"llparam.h"
#include	"llsymtab.h"
#include	<stdio.h>
#include	"llglobal.h"

int	rhs [MAXRHS];	/* Storage for the right hand side tokens */

int	rhsbase = 0;	/* The index of the most recent production */

int	rhstop = 0;	/* The index of the next available token */

int	*
mkrhsbase ()

{
	rhsbase = rhstop;
	return (&rhs [rhsbase]);
}

mkrhstoken (token)

{
	if (rhstop >= MAXRHS) {
		fprintf (stderr, "Too many production tokens!\n");
		exit (1);
	}

	rhs [rhstop++] = token;
}

rhsoffset ()

{
	return (rhstop - rhsbase - 1);
}

getrhstoken (offset)

int	offset;

{
	register	int	i;

	i = mkoffset (offset);

	return (rhs [i]);
}

setrhstoken (offset, token)

int	offset;
int	token;

{
	register	int	i;

	i = mkoffset (offset);
	rhs [i] = token;
}

mkoffset (offset)

int	offset;

{
	register	int	i;

	if (offset < 0 || (i = rhsbase + offset) >= rhstop) {
		fprintf (stderr, "Production token offset out of range!\n");
		exit (1);
	}

	return (i);
}

/*
 *	The following routine prints out each of the entries in
 *	the right hand side array.
 */

prrhs ()

{
	int	putrhs ();

	fprintf (tables, "\n/* The Right Hand Side Table */\n\n");
	fprintf (tables, "int llrhs [] = {\n");
	foreachprod (putrhs, 0);
	fprintf (tables, "};\n");
}

putrhs (n, lhs, rhs, rhslen, param)

int	n;
struct	symtab	*lhs;
int	rhs [];
int	rhslen;
int	param;

{
	register	int	i, j, k;
	register struct	symtab	*sp;

#ifdef	DEBUG
	fprintf (stderr, "Putrhs %d lhs %s length %d param %d.\n", n,
	lhs -> sname, rhslen, param);
#endif

	fprintf (tables, "\t");

	/*
	 *	First, dump out the left hand side attributes.
	 */

	j = lhs -> inherit + lhs -> synth;
	for (i = 0; i < j; i++) {
		fprintf (tables, "%d,", rhs [i]);
	}

	/*
	 *	Now, do the remainder of the symbols in the right
	 *	hand side, one at a time, along with their associated
	 *	attributes.
	 */

	while (i < rhslen) {
		sp = (struct symtab *) (rhs [i++]);
		j = sp -> inherit + sp -> synth;

		fprintf (tables, "%d,", sp -> snum);
		while (j--) {
			fprintf (tables, "%d,", rhs [i++]);
		}
	}
	fprintf (tables, "\n");
}

mkrhsindex (rhsaddr)

int	*rhsaddr;

{
	if (rhs <= rhsaddr && rhsaddr < &rhs [MAXRHS]) {
		return (rhsaddr - rhs);
	}

	return (-1);
}
