/*+
 *	Module:		llsymtab.c
 *
 *	Purpose:	Maintains the table of grammar symbols.
 *
 *	Invocation:	symptr = lookup (name);
 *
 *			trueorfalse = chksymattr (sym, inh, syn);
 *
 *			sortsymtab ();
 *
 *			foreachsym (routine, param);
 *
 *			symptr = getsymaddr (i);
 *
 *	Parameters:
 *
 *	  Input:	name - The address of the character string which
 *			is to be looked up in the table.  If the symbol is
 *			not already there, it is inserted.
 *
 *			sym - The address of a symbol table entry whose
 *			attributes are to be checked.
 *
 *			inh - The observed number of inherited attributes
 *			of the given symbol.
 *
 *			syn - The observed number of synthesized attributes
 *			of the given symbol.
 *
 *			routine - The address of a subroutine to be called
 *			once for each symbol table entry by foreachsym.
 *			The call sequence for routine is:
 *
 *				routine (n, sym, param);,
 *
 *			where n is the number of the symbol (counted from 0),
 *			sym is the address of the symbol table entry, and
 *			param is the parameter passed into foreachsym.
 *
 *			param - A parameter to be passed to the routine
 *			called by foreachsym.  Incredible liberties are
 *			taken with regard to the type of this parameter!
 *
 *			i - The sequence number of a symbol whose address
 *			is to be returned.  (Getsymaddr will only work
 *			properly AFTER symvec has been defined!)
 *
 *	  Output:	symptr - The address of the symbol table entry
 *			containing the named symbol.
 *
 *			trueorfalse - The boolean result of the consistency
 *			test for a symbol's attributes.
 *
 *	Author:		D. Bocking
 *
 *	Date:		16 October, 1984
-*/

#include	"llsymtab.h"
#include	"llparam.h"
#include	<stdio.h>
#include	"llglobal.h"

/*
 *	Storage for the symbol table proper, along with the address
 *	of the next available free table entry.
 */

struct	symtab	symtab [MAXSIZE];
struct	symtab	*symfree = symtab;

/*
 *	Storage for a vector of addresses of symbol table entries
 *	in declaration order by type.  This is generated by a call
 *	to sortsymtab ().
 */

struct	symtab	*symvec [MAXSIZE];

/*
 *	The hash table for rapid lookup.
 */

struct	symtab	*hash [NHASH];

/*
 *	The order of symbol type codes used in sorting the
 *	symbol table.
 */

int	symtypes [3] = {

	SYMRHS,
	SYMLHS,
	SYMACT

};

/*
 *	The following routine looks up an entry in the symbol table
 *	and returns a pointer to it.  An entry is added if necessary.
 */


struct symtab  *
lookup (name)

char	*name;

{
	int	hval, cmp;
	register	struct	symtab	*symptr, *succ, *t;
	register	char	*s;
	char	*symend, *savstring ();

#ifdef	DEBUG
	fprintf (stderr, "Lookup %s.\n", name);
#endif
	hval = 0;
	for (s = name; *s != '\0'; s++) {
		hval += *s & 0177;
	}
	hval %= NHASH;
#ifdef	DEBUG
	fprintf (stderr, "\tHash chain %d.\n", hval);
#endif

	symptr = hash [hval];
	t = (struct symtab *) 0;
	for ( ; symptr != (struct symtab *) 0; symptr = symptr -> chain) {
#ifdef	DEBUG
		fprintf (stderr, "\tComparing %s and %s.\n", symptr -> sname,
		name);
#endif
		if ((cmp = symcmp (symptr -> sname, name)) == 0) {
#ifdef	DEBUG
			fprintf (stderr, "\tFound symbol %s.\n", name);
#endif
			return (symptr);	/* symbol found. */
		}

		if (cmp > 0)
			break;		/* symbol not found. */
		t = symptr;
	}

	if ((s = savstring (name)) == (char *) 0) {
		fprintf (stderr, "%d: Symbol table overflow!\n", line);
		exit (1);
	}

	succ = symptr;
	symptr = symfree++;

	if (t == (struct symtab *) 0) {
		hash[hval] = symptr;
	} else {
		t -> chain = symptr;
	}

	symptr -> sname = s;
	symptr -> inherit = 0;
	symptr -> synth = 0;
	symptr -> stype = 0;
	symptr -> snum = 0;
	symptr -> chain = succ;

#ifdef	DEBUG
	fprintf (stderr, "\tNew symbol inserted.\n");
#endif
	return (symptr);
}

symcmp (as, at)
char   *as, *at;

{
	register char  *s, *t;

	s = as;
	t = at;
	for (; *s == *t; s++, t++)
	if (*s == '\0')
		return (0);
	return ((*s & 0377) < (*t & 0377) ? -1 : 1);
}

/*
 *	Define or validate the number of inherited and
 *	synthesized attributes for the current symbol.
 */

chksymattr (sym, inh, syn)

struct	symtab	*sym;
int	inh;
int	syn;

{
#ifdef	DEBUG
	fprintf (stderr, "Chksym %s %d %d.\n", sym -> sname, inh, syn);
#endif
	if (sym -> stype == 0) {
#ifdef	DEBUG
		fprintf (stderr, "\tSetting new values.\n");
#endif
		sym -> inherit = inh;
		sym -> synth = syn;
	} else {
		if (sym -> inherit != inh || sym -> synth != syn) {
			fprintf (stderr, "%d: %s %s.\n",
			line, "Inconsistent attribute usage in symbol --",
			sym -> sname);
			return (0);
		}
	}

	return (1);
}

/*
 *	The following routine produces the symvec list of addresses
 *	for the entries in the symbol table.  This vector is arranged
 *	in order by symbol type and declaration sequence.
 */

sortsymtab ()

{
	register	int	i, j, k;
	register struct	symtab	*sp;

	i = 0;
	for (j = 0; j < 3; j++) {
		k = symtypes [j];
		for (sp = symtab; sp < symfree; sp++) {
			if ((sp -> stype) == k) {
				symvec [i++] = sp;
			}
		}
	}

	for (; i < MAXSIZE; i++) {
		symvec [i] = (struct symtab *) 0;
	}
}

/*
 *	The following routine scans through the symbol table in
 *	sorted order, using symvec generated by sortsymtab.  (Thus,
 *	sortsymtab MUST be called prior to calls to this routine.)
 *	For each symbol in the table, a caller supplied subroutine is
 *	called with the symbol's number, the address of the symbol
 *	table entry, and a caller supplied parameter.
 */

foreachsym (routine, param)

int	(*routine) ();
int	param;

{
	register	int	i;

	for (i = 0; i < MAXSIZE && symvec [i] != (struct symtab *) 0; i++) {
		(*routine) (i, symvec [i], param);
	}
}

struct	symtab	*
getsymaddr (i)

int	i;

{
	if (i >= 0 && i < MAXSIZE) {
		return (symvec [i]);
	}

	return ((struct symtab *) 0);
}
